


import json
from typing import Dict, Mapping, Union


async def _write_cache_request(
    save_path: str, status_code: int, headers: Union[Dict[str, str], Mapping], content: bytes
) -> None:
    """
    Write the request's status code, headers, and content to a cache file.

    Args:
        head_path (str): The path to the cache file.
        status_code (int): The status code of the request.
        headers (Dict[str, str]): The dictionary of response headers.
        content (bytes): The content of the request.

    Returns:
        None
    """
    if not isinstance(headers, dict):
        headers = {k.lower():v for k, v in headers.items()}
    rq = {
        "status_code": status_code,
        "headers": headers,
        "content": content.hex(),
    }
    with open(save_path, "w", encoding="utf-8") as f:
        f.write(json.dumps(rq, ensure_ascii=False))


async def _read_cache_request(save_path: str) -> Dict[str, str]:
    """
    Read the request's status code, headers, and content from a cache file.

    Args:
        save_path (str): The path to the cache file.

    Returns:
        Dict[str, str]: A dictionary containing the status code, headers, and content of the request.
    """
    with open(save_path, "r", encoding="utf-8") as f:
        rq = json.loads(f.read())

    rq["content"] = bytes.fromhex(rq["content"])
    return rq



# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os
import shutil
import tempfile
from typing import Dict, Literal
from urllib.parse import urljoin
from fastapi import FastAPI, Request

import httpx
from olah.constants import CHUNK_SIZE, WORKER_API_TIMEOUT

from olah.utils.cache_utils import _read_cache_request, _write_cache_request
from olah.utils.rule_utils import check_cache_rules_hf
from olah.utils.repo_utils import get_org_repo
from olah.utils.file_utils import make_dirs


async def _tree_cache_generator(save_path: str):
    cache_rq = await _read_cache_request(save_path)
    yield cache_rq["headers"]
    yield cache_rq["content"]


async def tree_proxy_cache(
    app: FastAPI,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    commit: str,
    path: str,
    request: Request,
):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    # save
    method = request.method.lower()
    repos_path = app.app_settings.repos_path
    save_dir = os.path.join(
        repos_path, f"api/{repo_type}/{org}/{repo}/tree/{commit}/{path}"
    )
    save_path = os.path.join(save_dir, f"tree_{method}.json")

    # url
    org_repo = get_org_repo(org, repo)
    tree_url = urljoin(
        app.app_settings.config.hf_url_base(),
        f"/api/{repo_type}/{org_repo}/tree/{commit}/{path}",
    )
    headers = {}
    if "authorization" in request.headers:
        headers["authorization"] = request.headers["authorization"]
    async with httpx.AsyncClient() as client:
        response = await client.request(
            method=request.method,
            url=tree_url,
            headers=headers,
            timeout=WORKER_API_TIMEOUT,
            follow_redirects=True,
        )
        if response.status_code == 200:
            make_dirs(save_path)
            await _write_cache_request(
                save_path, response.status_code, response.headers, response.content
            )
        elif response.status_code == 404:
            pass
        else:
            raise Exception(
                f"Cannot get the branch info from the url {tree_url}, status: {response.status_code}"
            )


async def _tree_proxy_generator(
    app: FastAPI,
    headers: Dict[str, str],
    tree_url: str,
    allow_cache: bool,
    method: str,
    save_path: str,
):
    async with httpx.AsyncClient(follow_redirects=True) as client:
        content_chunks = []
        async with client.stream(
            method=method,
            url=tree_url,
            headers=headers,
            timeout=WORKER_API_TIMEOUT,
        ) as response:
            response_status_code = response.status_code
            response_headers = response.headers
            yield response_headers

            async for raw_chunk in response.aiter_raw():
                if not raw_chunk:
                    continue
                content_chunks.append(raw_chunk)
                yield raw_chunk

        content = bytearray()
        for chunk in content_chunks:
            content += chunk

        if response_status_code == 200:
            make_dirs(save_path)
            await _write_cache_request(
                save_path, response_status_code, response_headers, bytes(content)
            )


async def tree_generator(
    app: FastAPI,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    commit: str,
    path: str,
    request: Request,
):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    # save
    method = request.method.lower()
    repos_path = app.app_settings.repos_path
    save_dir = os.path.join(
        repos_path, f"api/{repo_type}/{org}/{repo}/tree/{commit}/{path}"
    )
    save_path = os.path.join(save_dir, f"tree_{method}.json")

    use_cache = os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    org_repo = get_org_repo(org, repo)
    tree_url = urljoin(
        app.app_settings.config.hf_url_base(),
        f"/api/{repo_type}/{org_repo}/tree/{commit}/{path}",
    )
    # proxy
    if use_cache:
        async for item in _tree_cache_generator(save_path):
            yield item
    else:
        async for item in _tree_proxy_generator(
            app, headers, tree_url, allow_cache, method, save_path
        ):
            yield item



# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from fastapi import Response
from fastapi.responses import JSONResponse


def error_repo_not_found() -> JSONResponse:
    return JSONResponse(
        content={"error": "Repository not found"},
        headers={
            "x-error-code": "RepoNotFound",
            "x-error-message": "Repository not found",
        },
        status_code=401,
    )


def error_page_not_found() -> Response:
    return Response(
        headers={
            "x-error-code": "RepoNotFound",
            "x-error-message": "Sorry, we can't find the page you are looking for.",
        },
        status_code=404,
    )

def error_entry_not_found(branch: str, path: str) -> Response:
    return Response(
        headers={
            "x-error-code": "EntryNotFound",
            "x-error-message": f"{path} does not exist on \"{branch}\"",
        },
        status_code=404,
    )




# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from asyncio import AbstractEventLoop
import json
import logging
import logging.handlers
import os
import platform
import re
import sys
from typing import AsyncGenerator, Generator
import warnings

import requests
import torch

from olah.constants import DEFAULT_LOGGER_DIR

handler = None


# Define a custom formatter without color codes
class NoColorFormatter(logging.Formatter):
    color_pattern = re.compile(r"\x1b[^m]*m")  # Regex pattern to match color codes

    def format(self, record):
        message = super().format(record)
        # Remove color codes from the log message
        message = self.color_pattern.sub("", message)
        return message


def build_logger(
    logger_name, logger_filename, logger_dir=DEFAULT_LOGGER_DIR
) -> logging.Logger:
    global handler

    formatter = logging.Formatter(
        fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    nocolor_formatter = NoColorFormatter(
        fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Set the format of root handlers
    if logging.getLogger().handlers is None or len(logging.getLogger().handlers) == 0:
        if sys.version_info[1] >= 9:
            # This is for windows
            logging.basicConfig(level=logging.INFO, encoding="utf-8")
        else:
            if platform.system() == "Windows":
                warnings.warn(
                    "If you are running on Windows, "
                    "we recommend you use Python >= 3.9 for UTF-8 encoding."
                )
            logging.basicConfig(level=logging.INFO)
    logging.getLogger().handlers[0].setFormatter(formatter)

    # Redirect stdout and stderr to loggers
    stdout_logger = logging.getLogger("stdout")
    stdout_logger.setLevel(logging.DEBUG)
    sl = StreamToLogger(stdout_logger, logging.INFO)
    sys.stdout = sl

    stderr_logger = logging.getLogger("stderr")
    stderr_logger.setLevel(logging.ERROR)
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    # Get logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    # Add a file handler for all loggers
    if handler is None:
        os.makedirs(logger_dir, exist_ok=True)
        filename = os.path.join(logger_dir, logger_filename)
        handler = logging.handlers.TimedRotatingFileHandler(
            filename, when="H", utc=True, encoding="utf-8"
        )
        handler.setFormatter(nocolor_formatter)
        handler.namer = lambda name: name.replace(".log", "") + ".log"

        for name, item in logging.root.manager.loggerDict.items():
            if isinstance(item, logging.Logger):
                item.addHandler(handler)

    return logger


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.terminal = sys.stdout
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def __getattr__(self, attr):
        try:
            attr_value = getattr(self.terminal, attr)
        except:
            return None
        return attr_value

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ""
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == "\n":
                encoded_message = line.encode("utf-8", "ignore").decode("utf-8")
                self.logger.log(self.log_level, encoded_message.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != "":
            encoded_message = self.linebuf.encode("utf-8", "ignore").decode("utf-8")
            self.logger.log(self.log_level, encoded_message.rstrip())
        self.linebuf = ""


def iter_over_async(
    async_gen: AsyncGenerator, event_loop: AbstractEventLoop
) -> Generator:
    """
    Convert async generator to sync generator

    :param async_gen: the AsyncGenerator to convert
    :param event_loop: the event loop to run on
    :returns: Sync generator
    """
    ait = async_gen.__aiter__()

    async def get_next():
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None

    while True:
        done, obj = event_loop.run_until_complete(get_next())
        if done:
            break
        yield obj



import argparse


import os
import sys
from olah.cache.olah_cache import OlahCache

def get_size_human(size: int) -> str:
    if size > 1024 * 1024 * 1024:
        return f"{int(size / (1024 * 1024 * 1024)):.4f}GB"
    elif size > 1024 * 1024:
        return f"{int(size / (1024 * 1024)):.4f}MB"
    elif size > 1024:
        return f"{int(size / (1024)):.4f}KB"
    else:
        return f"{size:.4f}B"

def insert_newlines(input_str, every=10):
    return '\n'.join(input_str[i:i+every] for i in range(0, len(input_str), every))

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Olah Cache Visualization Tool.")
    parser.add_argument("--file", "-f", type=str, required=True, help="The path of Olah cache file")
    parser.add_argument("--export", "-e", type=str, default="", help="Export the cached file if all blocks are cached")
    args = parser.parse_args()
    print(args)

    with open(args.file, "rb") as f:
        f.seek(0, os.SEEK_END)
        bin_size = f.tell()
    
    try:
        cache = OlahCache(args.file)
    except Exception as e:
        print(e)
        sys.exit(1)
    print(f"File: {args.file}")
    print(f"Olah Cache Version: {cache.header.version}")
    print(f"File Size: {get_size_human(cache.header.file_size)}")
    print(f"Cache Total Size: {get_size_human(bin_size)}")
    print(f"Block Size: {cache.header.block_size}")
    print(f"Block Number: {cache.header.block_number}")
    print(f"Cache Status: ")
    cache_status = cache.header.block_mask.__str__()[:cache.header._block_number]
    print(insert_newlines(cache_status, every=50))

    if args.export != "":
        if all([c == "1" for c in cache_status]):
            with open(args.file, "rb") as f:
                f.seek(cache._get_header_size(), os.SEEK_SET)
                with open(args.export, "wb") as fout:
                    fout.write(f.read())
        else:
            print("Some blocks are not cached, so the export is skipped.")


# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os
from peewee import *
import datetime

from olah.utils.olah_utils import get_olah_path



db_path = os.path.join(get_olah_path(), "database.db")
db = SqliteDatabase(db_path)


class BaseModel(Model):
    class Meta:
        database = db


class User(BaseModel):
    username = CharField(unique=True)



# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import platform
import os


def get_olah_path() -> str:
    if platform.system() == "Windows":
        olah_path = os.path.expanduser("~\\.olah")
    else:
        olah_path = os.path.expanduser("~/.olah")
    return olah_path
